#!/usr/bin/env python3

# Copyright © 2019-2020 Collabora Ltd.
#
# SPDX-License-Identifier: MIT
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

"""
Prepare external resources needed to build the Steam launcher .deb files.
"""

import argparse
import json
import logging
import os
import shutil
import subprocess
import tarfile
import tempfile
import urllib.request
from typing import (
    Any,
    Dict,
    List,
    Optional,
    Sequence,
)

from buildutils import SteamClient

logger = logging.getLogger('add-client-files')


BOOTSTRAP_RUNTIME_SONAMES = (
    'libX11-xcb.so.1',
    'libX11.so.6',
    'libXau.so.6',
    'libXdamage.so.1',
    'libXdmcp.so.6',
    'libXext.so.6',
    'libXfixes.so.3',
    'libXxf86vm.so.1',
    'libexpat.so.1',
    'libffi.so.6',
    'libgcc_s.so.1',
    'libstdc++.so.6',
    'libtinfo.so.5',
    'libxcb-dri2.so.0',
    'libxcb-dri3.so.0',
    'libxcb-glx.so.0',
    'libxcb-present.so.0',
    'libxcb-sync.so.1',
    'libxcb.so.1',
    'libz.so.1',
)


class InvocationError(Exception):
    pass


class Main:
    def __init__(
        self,
        client_manifest: str,
        client_uri: str,
        destination: str,
        runtime_snapshots_uri: str,
        beta_universe: bool = False,
        client_dir: Optional[str] = None,
        client_tarball_uri: Optional[str] = None,
        credential_envs: Sequence[str] = (),
        runtime_version: Optional[str] = None,
        **kwargs: Dict[str, Any],
    ) -> None:
        openers: List[urllib.request.BaseHandler] = []

        if credential_envs:
            password_manager = urllib.request.HTTPPasswordMgrWithDefaultRealm()

            for item in credential_envs:
                server, credential_env = item.split('=', 1)
                username, password = os.environ[credential_env].split(':', 1)

                password_manager.add_password(
                    # We ignore the realm. The type annotations say None
                    # isn't allowed here, but the documentation says it is!
                    None,       # type: ignore
                    f'https://{server}/',
                    username,
                    password,
                )

            openers.append(
                urllib.request.HTTPBasicAuthHandler(password_manager)
            )

        self.opener = urllib.request.build_opener(*openers)

        self.client_dir = None
        self.client_tarball_uri = client_tarball_uri
        self.client_manifest = client_manifest
        self.client_uri = client_uri
        self.client_version = None          # type: Optional[str]
        self.destination = destination
        self.resolved_runtime = None        # type: Optional[str]
        self.runtime_snapshots_uri = runtime_snapshots_uri
        self.runtime_version = runtime_version

        if beta_universe:
            self.package = 'steambeta'
        else:
            self.package = 'steam'

        if 'SOURCE_DATE_EPOCH' in os.environ:
            self.reference_timestamp = int(os.environ['SOURCE_DATE_EPOCH'])
        else:
            self.reference_timestamp = -1

    def fetch(
        self,
        uri: str,
        output: str,
    ) -> None:
        with self.opener.open(uri) as response:
            with open(output, 'wb') as writer:
                shutil.copyfileobj(response, writer)

    def get_runtime_uri(
        self,
        filename: str,
        version: str,
    ) -> str:
        return f'{self.runtime_snapshots_uri}/{version}/{filename}'

    def download_client(
        self,
        tmpdir: str,
    ) -> None:
        os.makedirs(os.path.join(tmpdir, 'client'), exist_ok=True)

        if self.client_tarball_uri is not None:
            self.fetch(
                self.client_tarball_uri,
                os.path.join(tmpdir, 'client.tar.gz'),
            )

            subprocess.run(
                [
                    'tar',
                    '--strip-components=1',
                    '-C', os.path.join(tmpdir, 'client'),
                    '-xvf', os.path.join(tmpdir, 'client.tar.gz'),
                ],
                check=True,
            )
        else:
            client = SteamClient()

            client.download_manifest(datadir=tmpdir)
            client.download_client(
                datadir=os.path.join(tmpdir, 'client'),
                strict=True,
            )
            runtimedir = os.path.join(tmpdir, 'client', 'ubuntu12_32')
            client.download_runtime(datadir=runtimedir, strict=True)
            client.extract_runtime(runtimedir=runtimedir, destdir=tmpdir)

            self.client_version = client.version
            self.resolved_runtime = client.runtime_version

        assert os.path.exists(
            os.path.join(tmpdir, 'client', 'steam.sh')
        )
        assert os.path.exists(
            os.path.join(tmpdir, 'client', 'ubuntu12_32', 'steam')
        )

    def ensure_scout_tarball(
        self,
        tmpdir: str,
    ) -> None:
        """
        Download a pre-prepared LD_LIBRARY_PATH Steam Runtime from a
        previous scout build.
        """
        path = os.path.join(
            tmpdir, 'client', 'ubuntu12_32', 'steam-runtime.tar.xz.part0',
        )

        if self.runtime_version is None:
            if not os.path.exists(path):
                raise InvocationError(
                    '--runtime-version must be specified if CLIENT_VERSION '
                    'does not contain a Steam Runtime tarball'
                )

            return

        with self.opener.open(
            self.get_runtime_uri(
                filename='VERSION.txt',
                version=self.runtime_version,
            )
        ) as response:
            resolved_runtime = response.read().strip().decode('utf-8')
            self.resolved_runtime = resolved_runtime

        logger.info(
            'Downloading steam-runtime build %s',
            resolved_runtime,
        )

        path = os.path.join(tmpdir, 'client', 'ubuntu12_32')

        os.makedirs(path, exist_ok=True)

        for filename in os.listdir(path):
            if (
                filename == 'steam-runtime.tar.xz'
                or filename.startswith('steam-runtime.tar.xz.part')
            ):
                os.remove(os.path.join(path, filename))

        filename = 'steam-runtime.tar.xz'

        self.fetch(
            self.get_runtime_uri(
                filename=filename,
                version=resolved_runtime,
            ),
            os.path.join(path, filename),
        )

        # The build script expects this to be split into parts.
        # Who are we to disagree? But for simplicity, we just use a
        # single part.
        os.link(
            os.path.join(path, filename),
            os.path.join(path, filename + '.part0'),
        )

    def install(
        self,
        src: str,
        dest: str,
        executable: bool = False,
    ) -> None:
        if executable:
            mode = '755'
        else:
            mode = '644'

        subprocess.run([
            'install', '-p', '-m', mode, src, dest,
        ], check=True)

    def _normalize_tar_entry(
        self,
        entry: tarfile.TarInfo,
    ) -> tarfile.TarInfo:
        entry.uid = 65534
        entry.gid = 65534

        if (
            self.reference_timestamp != -1
            and entry.mtime > self.reference_timestamp
        ):
            entry.mtime = self.reference_timestamp

        entry.uname = 'nobody'
        entry.gname = 'nogroup'

        return entry

    def build_bootstrap(
        self,
        client_dir: str,
        tmpdir: str,
    ) -> None:
        os.makedirs(
            os.path.join(tmpdir, 'bootstrap', 'linux32'),
            exist_ok=True,
            mode=0o755,
        )
        os.makedirs(
            os.path.join(tmpdir, 'bootstrap', 'ubuntu12_32'),
            exist_ok=True,
            mode=0o755,
        )

        self.install(
            os.path.join(client_dir, 'linux32', 'steamerrorreporter'),
            os.path.join(tmpdir, 'bootstrap', 'linux32', ''),
            executable=True,
        )
        self.install(
            os.path.join(client_dir, 'steam.sh'),
            os.path.join(tmpdir, 'bootstrap', ''),
            executable=True,
        )
        self.install(
            os.path.join(client_dir, 'steamdeps.txt'),
            os.path.join(tmpdir, 'bootstrap', ''),
            executable=False,
        )
        self.install(
            os.path.join(client_dir, 'ubuntu12_32', 'steam'),
            os.path.join(tmpdir, 'bootstrap', 'ubuntu12_32', ''),
            executable=True,
        )
        self.install(
            os.path.join(client_dir, 'ubuntu12_32', 'crashhandler.so'),
            os.path.join(tmpdir, 'bootstrap', 'ubuntu12_32', ''),
            executable=True,
        )

        runtimedir = os.path.join(client_dir, 'ubuntu12_32')
        bootstrap = os.path.join(tmpdir, 'bootstrap')
        bootstrap_runtime_dir = os.path.join(bootstrap, 'ubuntu12_32')

        with tarfile.open(
            os.path.join(runtimedir, 'steam-runtime.tar.xz'), mode='r|xz',
        ) as tar_reader:
            for info in tar_reader:
                bits = info.name.split('/')

                if bits[0] != 'steam-runtime':
                    raise ValueError(f'{info.name} is not in steam-runtime/')

                if '..' in bits:
                    raise ValueError(f'{info.name} has path traversal')

                if bits[1] in (
                    'COPYING',
                    'README.txt',
                    'built-using.txt',
                    'common-licenses',
                    'manifest.deb822.gz',
                    'manifest.txt',
                    'run.sh',
                    'scripts',
                    'setup.sh',
                    'version.txt',
                ):
                    tar_reader.extract(info, path=bootstrap_runtime_dir)
                    continue

                if bits[1] == 'usr':
                    bits = bits[2:]
                else:
                    bits = bits[1:]

                if bits[0:2] != ['lib', 'i386-linux-gnu']:
                    continue

                if len(bits) != 3:
                    continue

                for soname in BOOTSTRAP_RUNTIME_SONAMES:
                    if bits[2] == soname or bits[2].startswith(soname + '.'):
                        tar_reader.extract(info, path=bootstrap_runtime_dir)
                        break

        with tarfile.open(
            os.path.join(
                self.destination,
                'bootstraplinux_ubuntu12_32.tar.xz',
            ),
            mode='w|xz',
            format=tarfile.GNU_FORMAT,
        ) as tar_writer:
            members = []

            for dir_path, dirs, files in os.walk(
                bootstrap,
                topdown=True,
                followlinks=False,
            ):
                rel_dir_path = os.path.relpath(dir_path, bootstrap)

                for member in dirs + files:
                    if rel_dir_path == '.':
                        members.append(member)
                    else:
                        members.append(os.path.join(rel_dir_path, member))

            for member in sorted(members):
                tar_writer.add(
                    os.path.join(bootstrap, member),
                    arcname=member,
                    recursive=False,
                    filter=self._normalize_tar_entry,
                )

    def run(self) -> None:
        with tempfile.TemporaryDirectory() as tmpdir:
            if self.client_dir is not None:
                client_dir = self.client_dir
            else:
                client_dir = os.path.join(tmpdir, 'client')
                self.download_client(tmpdir)
                self.ensure_scout_tarball(tmpdir)

            self.build_bootstrap(client_dir, tmpdir)

        info = {
            'client_version': self.client_version,
            'runtime_version': self.resolved_runtime,
        }
        with open(
            os.path.join(self.destination, 'client-versions.json'),
            'w'
        ) as writer:
            json.dump(info, writer, indent=2)
            writer.write('\n')


def main() -> None:
    logging.basicConfig()
    logging.getLogger().setLevel(logging.DEBUG)

    parser = argparse.ArgumentParser(
        description=__doc__,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    parser.add_argument(
        '--beta-universe', default=False,
        help='Build a steambeta package',
    )
    parser.add_argument(
        '--client-dir', default=None,
        help=(
            'Use a pre-downloaded Steam Client instead of downloading it '
            'from the Steam CDN'
        ),
    )
    parser.add_argument(
        '--client-uri', default='https://steamcdn-a.akamaihd.net/client',
        help='Base URI for Steam client files',
    )
    parser.add_argument(
        '--client-manifest', default='steam_client_ubuntu12',
        help='Client manifest VDF file relative to CLIENT_URI',
    )
    parser.add_argument(
        '--client-tarball-uri', default=None,
        help=(
            'Download Steam Client files from this URI instead of from '
            'the Steam client CDN'
        ),
    )
    parser.add_argument(
        '--credential-env', dest='credential_envs', action='append',
        default=[], metavar='HOSTNAME=VARIABLE',
        help=(
            'Evaluate environment variable VARIABLE to get the '
            'username:password to use for https://HOSTNAME, '
            'for example `export SERVER_CREDS=gfreeman:n1h1l4nth` '
            'and use --credential-env=server.example.com=SERVER_CREDS'
        ),
    )
    parser.add_argument(
        '--runtime-snapshots-uri',
        default=(
            'https://repo.steampowered.com/steamrt-images-scout/snapshots'
        ),
        help=(
            'Download Steam Runtime tarball from a subdirectory of this'
        ),
    )
    parser.add_argument(
        '--runtime-version', default=None,
        help='Replace Steam Runtime tarball (if any) with this version',
    )

    parser.add_argument(
        '--destination',
        default=os.path.dirname(os.path.dirname(__file__)),
        help='Directory containing Steam launcher deb packaging',
    )

    try:
        args = parser.parse_args()
        Main(**vars(args)).run()
    except InvocationError as e:
        parser.error(str(e))


if __name__ == '__main__':
    main()



# Copyright © 2020 Collabora Ltd.
#
# SPDX-License-Identifier: MIT
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import datetime
import hashlib
import io
import logging
import os
import shutil
import sys
import tarfile
import tempfile
import typing
import urllib.error
import urllib.request
import zipfile
from contextlib import suppress

# git remote add --no-tags python-vdf https://github.com/ValvePython/vdf
# Update with:
# git subtree merge -P subprojects/python-vdf python-vdf/master
sys.path[:0] = [
    os.path.join(
        os.path.dirname(os.path.dirname(__file__)),
        'subprojects',
        'python-vdf'
    ),
]

import vdf      # noqa


logger = logging.getLogger('buildutils-lib')


class HashingWriter(io.BufferedIOBase):
    def __init__(self, name, mode):     # type: (str, str) -> None
        self._writer = open(name, mode)
        self._sha256 = hashlib.sha256()
        self.size = 0

    def write(self, blob):  # type: (bytes,) -> int
        self.size += len(blob)
        self._sha256.update(blob)
        self._writer.write(blob)
        return len(blob)

    @property
    def sha256(self):   # type: () -> str
        return self._sha256.hexdigest()

    def __enter__(self):    # type: () -> HashingWriter
        return self

    def __exit__(self, *rest):
        return self._writer.__exit__(*rest)


def verbose_urlopen(
    url     # type: str
):
    # type: (...) -> typing.BinaryIO

    logger.info('Requesting <%s>...', url)
    try:
        return urllib.request.urlopen(url)
    except urllib.error.URLError:
        logger.error('Error opening <%s>:', url)
        raise


class SteamClient:
    def __init__(
        self,
        *,
        uri: str = 'https://steamcdn-a.akamaihd.net/client',
        manifest: str = 'steam_client_ubuntu12'
    ):
        self.uri = uri
        self.manifest = manifest
        self.version = None         # type: typing.Optional[str]
        self.datetime = None        # type: typing.Optional[datetime.datetime]
        self.manifest_content = {
        }       # type: typing.Dict[str, typing.Dict[str, typing.Any]]
        self.have_runtime_manifest = False
        self.runtime_version = None             # type: typing.Optional[str]
        self.runtime_version_marker = None      # type: typing.Optional[str]

    def download_manifest(
        self,
        datadir: str
    ) -> None:

        os.makedirs(datadir, exist_ok=True)

        logger.info(
            'Downloading Steam client manifest from %s/%s...',
            self.uri, self.manifest,
        )
        with verbose_urlopen(
            '{}/{}'.format(self.uri, self.manifest)
        ) as http_reader:
            with open(
                os.path.join(datadir, 'manifest.vdf'), 'wb'
            ) as writer:
                shutil.copyfileobj(http_reader, writer)

        self.load_manifest(datadir)

    def download_client(
        self,
        datadir: str,
        strict: bool = True,
    ) -> None:
        names = []      # type: typing.List[str]

        with tempfile.TemporaryDirectory() as tempdir:
            for name, value in sorted(
                self.manifest_content['ubuntu12'].items(),
                key=lambda i: i[0]
            ):
                if name == 'version':
                    continue

                basename = value['file']
                assert '/' not in basename, basename
                names.append(basename)
                with verbose_urlopen(
                    '{}/{}'.format(self.uri, basename)
                ) as downloader:
                    with HashingWriter(
                        os.path.join(tempdir, basename),
                        'wb'
                    ) as hasher:
                        shutil.copyfileobj(downloader, hasher)
                        if hasher.sha256 != value['sha2']:
                            logger.warning(
                                'Unexpected hash for %s/%s\n'
                                '  Expected: %s\n'
                                '  Got     : %s\n',
                                self.uri, basename,
                                value['sha2'], hasher.sha256)

                            if strict:
                                raise ValueError('sha256 mismatch')

                        if hasher.size != int(value['size']):
                            logger.warning(
                                'Unexpected size for %s/%s\n'
                                '  Expected: %s\n'
                                '  Got     : %s\n',
                                self.uri, basename,
                                value['size'], hasher.size)

                            if strict:
                                raise ValueError('sha256 mismatch')

            for basename in names:
                with zipfile.ZipFile(
                    os.path.join(tempdir, basename), 'r'
                ) as unzip:
                    for part in unzip.infolist():
                        part.filename = part.filename.replace('\\', '/')
                        unzip.extract(part, path=datadir)

    def download_runtime(
        self,
        datadir: str,
        strict: bool = False,
    ) -> None:
        names = []      # type: typing.List[str]

        with tempfile.TemporaryDirectory() as tempdir:
            for name, value in sorted(
                self.manifest_content['ubuntu12'].items(),
                key=lambda i: i[0]
            ):
                if (name.startswith('runtime_part')
                        and name.endswith('_ubuntu12')):
                    basename = value['file']
                    assert '/' not in basename, basename
                    names.append(basename)
                    with verbose_urlopen(
                        '{}/{}'.format(self.uri, basename)
                    ) as downloader:
                        with HashingWriter(
                            os.path.join(tempdir, basename),
                            'wb'
                        ) as hasher:
                            shutil.copyfileobj(downloader, hasher)
                            if hasher.sha256 != value['sha2']:
                                logger.warning(
                                    'Unexpected hash for %s/%s\n'
                                    '  Expected: %s\n'
                                    '  Got     : %s\n',
                                    self.uri, basename,
                                    value['sha2'], hasher.sha256)

                                if strict:
                                    raise ValueError('sha256 mismatch')

                            if hasher.size != int(value['size']):
                                logger.warning(
                                    'Unexpected size for %s/%s\n'
                                    '  Expected: %s\n'
                                    '  Got     : %s\n',
                                    self.uri, basename,
                                    value['size'], hasher.size)

                                if strict:
                                    raise ValueError('sha256 mismatch')

            with open(
                os.path.join(datadir, 'steam-runtime.tar.xz'), 'wb'
            ) as tar_writer:
                for basename in names:
                    with zipfile.ZipFile(
                        os.path.join(tempdir, basename), 'r'
                    ) as unzip:
                        for part in unzip.infolist():
                            if '.tar.xz.part' in part.filename:
                                with unzip.open(part) as part_reader:
                                    shutil.copyfileobj(
                                        part_reader, tar_writer)

    def load_manifest(
        self,
        datadir: str,
    ) -> None:
        with open(
            os.path.join(datadir, 'manifest.vdf'), 'r'
        ) as text_reader:
            self.manifest_content = vdf.load(text_reader)

        self.version = str(
            self.manifest_content['ubuntu12']['version']
        )

        logger.info(
            'Steam client build: %s', self.version,
        )

        try:
            timestamp = int(self.version)
        except ValueError:
            self.datetime = None
        else:
            self.datetime = datetime.datetime.fromtimestamp(
                timestamp,
                datetime.timezone.utc,
            )
            logger.info(
                'Steam client build date (probably) %s',
                self.datetime.strftime('%Y-%m-%d %H:%M:%S%z'),
            )

    def extract_runtime(
        self,
        runtimedir: str,
        destdir: str,
        extract_manifest: bool = False,
    ) -> None:

        with suppress(FileNotFoundError):
            shutil.rmtree(os.path.join(runtimedir, 'steam-runtime'))

        with suppress(FileNotFoundError):
            shutil.rmtree(os.path.join(destdir, 'steam-runtime'))

        with suppress(FileExistsError):
            os.mkdir(os.path.join(runtimedir, 'steam-runtime'))

        with suppress(FileExistsError):
            os.mkdir(os.path.join(destdir, 'steam-runtime'))

        with tarfile.open(
            os.path.join(runtimedir, 'steam-runtime.tar.xz'), mode='r|xz',
        ) as tar_reader:
            for info in tar_reader:
                if not info.isfile():
                    continue

                if info.name in (
                    'steam-runtime/version.txt',
                ):
                    tar_reader.extract(info, path=destdir)
                    continue

                if extract_manifest and info.name in (
                    'steam-runtime/built-using.txt',
                    'steam-runtime/manifest.deb822.gz',
                    'steam-runtime/manifest.txt',
                ):
                    tar_reader.extract(info, path=destdir)

                    if info.name == 'steam-runtime/manifest.deb822.gz':
                        self.have_runtime_manifest = True

                    continue

        with suppress(
            FileNotFoundError
        ), open(
            os.path.join(
                destdir, 'steam-runtime', 'version.txt'),
            'r'
        ) as text_reader:
            marker = text_reader.read().strip()

        self.runtime_version_marker = marker
        version_bits = marker.split('_')

        if len(version_bits) != 2 or version_bits[0] != 'steam-runtime':
            logger.warning(
                'Unexpected format for runtime version: %s', marker,
            )
        else:
            self.runtime_version = version_bits[-1]
            logger.info('Runtime version %s', self.runtime_version)



#!/usr/bin/env python3
"""
    This script handles installing system dependencies for games using the
    Steam runtime.  It is intended to be customized by other distributions
    to "do the right thing"

    Usage: steamdeps dependencies.txt
"""

import argparse
import os
import re
import shlex
import stat
import subprocess
import sys
import tempfile

try:
    import typing
except ImportError:
    pass

try:
    import apt
except ImportError:
    sys.stderr.write("Couldn't import apt, please install python3-apt or "
                     "update steamdeps for your distribution.\n")
    sys.exit(3)

# This is the set of supported Steam runtime environments
SUPPORTED_STEAM_RUNTIME = ['1']

# This is the set of supported dependency formats
SUPPORTED_STEAM_DEPENDENCY_VERSION = ['1']

_arch = None


class OsRelease:
    def __init__(self):
        # type: () -> None
        self._load_any()
        self._is_like = self._data.get('ID_LIKE', '').split()

    def _load_any(self):
        # type: () -> None
        for path in ('/etc/os-release', '/usr/lib/os-release'):
            self._data = self._load(path)

            if self._data:
                return

    def _load(
        self,
        path        # type: str
    ):
        # type: (...) -> typing.Dict[str, str]

        data = {}       # type: typing.Dict[str, str]

        try:
            with open(path, 'r', encoding='utf-8', errors='replace') as reader:
                for line in reader:
                    try:
                        line = line.strip()

                        if not line or line.startswith('#'):
                            continue

                        key, value = line.split('=', 1)

                        data[key] = ''.join(shlex.split(value))

                    except Exception as e:
                        print(
                            'Warning: error parsing %s: line %r: %s'
                            % (path, line, e),
                            file=sys.stderr
                        )
                        continue
        except OSError:
            return {}
        else:
            return data

    def is_os(self, os_id):
        # type: (str) -> bool
        if self._data.get('ID') == os_id:
            return True

        return os_id in self._is_like

    def dump(self):
        print(self._data, file=sys.stderr)


#
# Get the current package architecture
# This may be different than the actual architecture for the case of i386
# chroot environments on amd64 hosts.
#
def get_arch():
    """
    Get the current architecture
    """
    global _arch

    if _arch is None:
        _arch = subprocess.check_output(
            ['dpkg', '--print-architecture']).decode("utf-8").strip()
    return _arch


###
def get_full_package_name(name):
    """
    Get the full name of a package, qualified by architecture
    """
    if name.find(":") < 0:
        return name + ":" + get_arch()
    else:
        return name


# N.B. Version checks are not supported on virtual packages
#
def is_provided(pkgname):
    """
    Check to see if another package Provides this package
    """
    cache = apt.Cache()
    pkgs = cache.get_providing_packages(pkgname)
    for pkg in pkgs:
        if pkg.is_installed:
            return True
    return False


###
class Package:
    """
    Package definition class
    """

    def __init__(self, name, version_conditions):
        self.name = name
        self.version_conditions = version_conditions
        self.installed = None

    def set_installed(self, version):
        self.installed = version

    def is_available(self):
        if self.installed is None:
            # check to see if another package is providing this virtual package
            return is_provided(self.name)

        for (op, version) in self.version_conditions:
            if subprocess.call(['dpkg', '--compare-versions', self.installed,
                                op, version]) != 0:
                return False

        return True

    def __str__(self):
        text = self.name
        for (op, version) in self.version_conditions:
            text += " (%s %s)" % (op, version)
        return text


def is_glvnd():
    try:
        with subprocess.Popen(['apt-cache', 'pkgnames', 'libgl1'],
                              stdout=subprocess.PIPE,) as process:
            for line in process.stdout:
                line = line.decode('utf-8').strip()

                if line == 'libgl1':
                    return True
            return False
    except (OSError, FileNotFoundError):
        return False


def is_package_available(package_name):
    try:
        apt_output = subprocess.check_output(
            ['apt-cache', '-q', 'show', package_name])
        # On SteamOS "xdg-desktop-portal-gtk" is recommended by the backported
        # version of flatpak, even if it is not available in the repositories.
        # For this reason we also check the apt-cache -q show output before
        # asserting that a package is really available.
        return bool(apt_output)
    except (OSError, FileNotFoundError, subprocess.CalledProcessError):
        return False


def choose_xdg_portal_backend(default_backend="xdg-desktop-portal-gtk"):
    gtk_desktop = ["Cinnamon", "GNOME", "LXDE", "MATE", "Unity", "XFCE"]
    kde_desktop = ["KDE"]
    gtk_backend = "xdg-desktop-portal-gtk"
    kde_backend = "xdg-desktop-portal-kde"

    current_desktops = os.environ.get("XDG_CURRENT_DESKTOP", "").split(":")

    for c in current_desktops:
        if c in gtk_desktop:
            if is_package_available(gtk_backend):
                return gtk_backend
        elif c in kde_desktop:
            if is_package_available(kde_backend):
                return kde_backend

    if is_package_available(default_backend):
        return default_backend
    else:
        sys.stderr.write(
            "There isn't a known XDG portal backend that can be installed\n")
        return None


def remap_package(name):
    if name in (
            'python-apt',
    ):
        # Steam claims it needs python-apt, but it doesn't really
        return None

    # Ubuntu 12.04.2, 12.04.3, and 12.04.4 introduce new X stacks which require
    # different sets of incompatible glx packages depending on which X
    # is currently installed.

    cache = apt.Cache()
    for lts in ('quantal', 'raring', 'saucy', 'trusty', 'xenial'):
        xserver = 'xserver-xorg-core-lts-' + lts
        if xserver in cache and cache[xserver].is_installed:
            if name in (
                    'libgl1-mesa-glx',
                    'libgl1-mesa-dri',
            ):
                return name + '-lts-' + lts

    if name == 'libgl1-mesa-glx':
        if is_glvnd():
            return 'libgl1'

    return name


###
def create_package(description):
    """
    Create a package object based on a description.
    This can return None if the package isn't meaningful on this platform.
    """
    # Look for architecture conditions, e.g. foo [i386]
    match = re.match(r"(.*) \[([^\]]+)\]", description)
    if match is not None:
        description = match.group(1).strip()
        condition = match.group(2)
        if condition[0] == '!':
            if get_arch() == condition[1:]:
                return None
        else:
            if get_arch() != condition:
                return None

    # Look for version requirements, e.g. foo (>= 1.0)
    version_conditions = []
    while True:
        match = re.search(r"\s*\(\s*([<>=]+)\s*([\w\-.:]+)\s*\)\s*",
                          description)
        if match is None:
            break

        version_conditions.append((match.group(1), match.group(2)))
        description = description[:match.start()] + description[match.end():]

    description = description.strip()

    if ':' in description:
        name, multiarch = description.rsplit(':', 1)
    else:
        name = description
        multiarch = None

    name = remap_package(name)

    if name in (
            "xdg-desktop-portal-gtk",
            "xdg-desktop-portal-kde",
    ):
        name = choose_xdg_portal_backend(name)
        # Skip version conditions for portal backends because we don't know
        # in advance which backend we will need to install.
        # In the future, if necessary, we can enhance this check and, for
        # example, embed a minimum version for each possible backed.
        version_conditions = []

    if name is None:
        return None
    elif multiarch is not None:
        description = name + ':' + multiarch
    else:
        description = name

    return Package(description, version_conditions)


###
def get_terminal_command_line(title):
    """
    Function to find a useful terminal like xterm or compatible
    """
    if "DISPLAY" in os.environ:
        gnome_wait_option = None
        try:
            # Use the new '--wait' option if available
            terminal_out = subprocess.check_output(
                ["gnome-terminal", "--help-terminal-options"]).decode("utf-8")
            if "--wait" in terminal_out:
                gnome_wait_option = "--wait"
            else:
                # If the old '--disable-factory' is supported we use it
                terminal_out = subprocess.check_output(
                    ["gnome-terminal", "--help"]).decode("utf-8")
                if "--disable-factory" in terminal_out:
                    gnome_wait_option = "--disable-factory"

            if gnome_wait_option is not None:
                # If 'gnome-terminal' with the right options is available, we
                # just use it
                return ["gnome-terminal", gnome_wait_option, "-t", title, "--"]
        except FileNotFoundError:
            pass

        programs = [
            ("konsole",
             ["konsole", "--nofork", "-p", "tabtitle=" + title, "-e"]),
            ("xterm",
             ["xterm", "-bg", "#383635", "-fg", "#d1cfcd", "-T", title, "-e"]),
            ("x-terminal-emulator",
             ["x-terminal-emulator", "-T", title, "-e"]),
            # If we reach this point either 'gnome-terminal' is not available
            # or the current version is too old for the new '--wait' option.
            # Anyway we can't know for sure if '--disable-factory' option
            # is supported until we try it because, for example,
            # on Ubuntu 16.04 '--disable-factory' is available but it doesn't
            # show up with '--help'. Leave this 'gnome-terminal' test as the
            # last resort because it's highly likely to fail.
            ("gnome-terminal",
             ["gnome-terminal", "--disable-factory", "-t", title, "--"]),
        ]
        for (program, commandLine) in programs:
            if subprocess.call(['which', program],
                               stdout=subprocess.PIPE) == 0:
                return commandLine

    # Fallback if no GUI terminal program is available
    return ['/bin/sh']


###
def update_packages(packages):
    """
    Function to install or update package dependencies
    Ideally we would call some sort of system UI that users were familiar with
    to do this, but nothing that exists yet does what we need.
    """

    package_list = " ".join([package.name for package in packages])

    # Create a temporary file to hold the installation completion status
    (fd, status_file) = tempfile.mkstemp()
    os.close(fd)

    # Create a script to run, in a secure way
    (fd, script_file) = tempfile.mkstemp()
    script = """#!/bin/sh
check_sudo()
{
    # If your host file is misconfigured in certain circumstances this
    # can cause sudo to block for a while, which causes gksudo to go into
    # limbo and never return.
    timeout --signal=9 5 sudo -v -S </dev/null 2>/dev/null
    if [ $? -eq 124 -o $? -eq 137 ]; then
        # sudo timed out or was killed due to timeout
        cat <<__EOF__
sudo timed out, your hostname may be missing from /etc/hosts.

See https://support.steampowered.com/kb_article.php?ref=7493-ADXN-9620
for more details.
__EOF__
        return 1
    else
        return 0
    fi
}

cat <<__EOF__
Steam needs to install these additional packages:
    %s
__EOF__
check_sudo

# Check to make sure 64-bit systems can get 32-bit packages
if [ "$(dpkg --print-architecture)" = "amd64" ] && \
   ! dpkg --print-foreign-architectures | grep i386 >/dev/null; then
    sudo dpkg --add-architecture i386
fi

# Update the package list, showing progress
sudo apt-get update | while read line; do echo -n "."; done
echo

# Install the packages!
sudo apt-get install %s
echo $? >%s
echo -n "Press return to continue: "
read line
""" % (", ".join([package.name for package in packages]), package_list,
       status_file)
    os.write(fd, script.encode("utf-8"))
    os.close(fd)
    os.chmod(script_file, (stat.S_IRUSR | stat.S_IWUSR | stat.S_IXUSR))

    try:
        subprocess.call(
            get_terminal_command_line("Package Install") + [script_file])
    except KeyboardInterrupt:
        pass
    os.unlink(script_file)

    # Read the status out of the file, since if we ran the script in a
    # terminal the process status will be whether the terminal started
    try:
        status = int(open(status_file).read())
    except ValueError:
        # The status wasn't written to the file
        status = 255

    os.unlink(status_file)

    return status


###
def check_config(path, config):
    if "STEAM_RUNTIME" not in config:
        sys.stderr.write(
            "Missing STEAM_RUNTIME definition in %s\n" % path)
        return False

    if config["STEAM_RUNTIME"] not in SUPPORTED_STEAM_RUNTIME:
        sys.stderr.write(
            "Unsupported Steam runtime: %s\n" % config["STEAM_RUNTIME"])
        return False

    if "STEAM_DEPENDENCY_VERSION" not in config:
        sys.stderr.write(
            "Missing STEAM_DEPENDENCY_VERSION definition in %s\n" % path)
        return False

    if config["STEAM_DEPENDENCY_VERSION"]\
            not in SUPPORTED_STEAM_DEPENDENCY_VERSION:
        sys.stderr.write("Unsupported dependency version: %s\n" % config[
            "STEAM_DEPENDENCY_VERSION"])
        return False

    # Make sure we can use dpkg on this system.
    try:
        subprocess.call(['dpkg', '--version'], stdout=subprocess.PIPE)
    except FileNotFoundError:
        sys.stderr.write("Couldn't find dpkg, please update steamdeps for "
                         "your distribution.\n")
        return False

    return True


###
def main():
    config = {}
    os_release = OsRelease()
    missing_packages = []    # type: List[str]

    parser = argparse.ArgumentParser(description='Install Steam dependencies')
    parser.add_argument(
        '--dry-run',
        action='store_true',
        help="Don't install anything, just report",
    )
    parser.add_argument(
        '--debug-dump-os-release',
        action='store_true',
        help=argparse.SUPPRESS,
    )
    parser.add_argument(
        'dependencies',
        metavar='$HOME/.steam/root/steamdeps.txt',
        help='Path to steamdeps.txt',
    )
    args = parser.parse_args()

    if args.debug_dump_os_release:
        os_release.dump()
        return 0

    # Make sure we can open the file
    try:
        fp = open(args.dependencies)
    except Exception as e:
        sys.stderr.write("Couldn't open file: %s\n" % e)
        return 2

    # Look for configuration variables
    config_pattern = re.compile(r"(\w+)\s*=\s*(\w+)")
    for line in fp:
        line = line.strip()
        if line == "" or line[0] == '#':
            continue

        match = re.match(config_pattern, line)
        if match is not None:
            config[match.group(1)] = match.group(2)

    # Check to make sure we have a valid config
    if not check_config(args.dependencies, config):
        return 3

    # Seek back to the beginning of the file
    fp.seek(0)

    # Load the package dependency information
    packages = {}
    dependencies = []
    for line in fp:
        line = line.strip()
        if line == "" or line[0] == '#':
            continue

        match = re.match(config_pattern, line)
        if match is not None:
            continue

        row = []
        for section in line.split("|"):
            package = create_package(section)
            if package is None:
                continue

            packages[package.name] = package
            row.append(package)

        dependencies.append(row)

    if get_arch() == 'amd64':
        for synthetic in (
                'libc6',
                'libgl1-mesa-dri',
                'libgl1-mesa-glx',
        ):
            package = create_package(synthetic + ':amd64')
            if package is not None:
                packages[package.name] = package
                dependencies.append([package])

    # Try to install these packages, even if they are not
    # listed in the steamdeps.txt file. If they are not available we
    # just inform the user about it and continue.
    for additional_pkg in (
        'steam-libs-amd64:amd64',
        'steam-libs-i386:i386',
        'xdg-desktop-portal',
        choose_xdg_portal_backend() or 'xdg-desktop-portal-gtk',
    ):
        if additional_pkg not in packages:
            if is_package_available(additional_pkg):
                package = Package(additional_pkg, [])
                packages[package.name] = package
                dependencies.append([package])
            else:
                missing_packages.append(additional_pkg)

    # The Steam container runtime (pressure-vessel) requires a setuid
    # bubblewrap executable on some kernel configurations. Steam is
    # unprivileged, so we have to get it from the host OS.
    if (
        # Debian's kernel doesn't allow unprivileged users to create
        # new namespaces (https://bugs.debian.org/898446) so we need the
        # setuid bubblewrap
        os_release.is_os('debian')
        # Ubuntu's kernel does allow that. We assume Ubuntu derivatives
        # like Linux Mint will inherit that, rather than reverting to the
        # Debian behaviour
        and not os_release.is_os('ubuntu')
    ):
        package = Package('bubblewrap', [])
        packages[package.name] = package
        dependencies.append([package])

    if missing_packages:
        print("These packages are not available:\n")

        for p in sorted(missing_packages):
            print("- %s" % p)

        print("\nThe Steam client may have limited functionality.")

    # Print package dependency information for debug
    """
    for row in dependencies:
        print " | ".join( [ str(package) for package in row ] )
    """

    # Get the installed package versions
    # Make sure COLUMNS isn't set, or dpkg will truncate its output
    if "COLUMNS" in os.environ:
        del os.environ["COLUMNS"]

    process = subprocess.Popen(['dpkg', '-l'] + list(packages.keys()),
                               stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    installed_pattern = re.compile(r"^\Si\s+([^\s]+)\s+([^\s]+)")
    for line in process.stdout:
        line = line.decode("utf-8").strip()
        match = re.match(installed_pattern, line)
        if match is None:
            continue

        name = match.group(1)
        if name not in packages:
            name = get_full_package_name(name)
        packages[name].set_installed(match.group(2))

    # See which ones need to be installed
    needed = []
    for row in dependencies:
        if len(row) == 0:
            continue

        satisfied = False
        for dep in row:
            if dep.is_available():
                satisfied = True
                break
        if not satisfied:
            needed.append(row[0])

    # If we have anything to install, do it!
    if len(needed) > 0:
        for package in needed:
            if package.installed:
                print("Package %s is installed with version '%s' but doesn't "
                      "match requirements: %s" % (
                        package.name, package.installed, package),
                      file=sys.stderr)
            else:
                print("Package %s needs to be installed" % package.name,
                      file=sys.stderr)

        if args.dry_run:
            return 1
        else:
            return update_packages(needed)
    else:
        return 0


if __name__ == "__main__":
    sys.exit(main())
